
/* ISX to VB ROM Converter v1.0
Copyright (C) 2010 Amos "RunnerPack" Bieler.  All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

typedef unsigned long u32;
typedef signed long s32;
typedef unsigned char u8;

// Calculate the next larger power of two
u32 ceilp2(u32 x) {
	u32 y = x - 1;
	y = y | (y >> 1);
	y = y | (y >> 2);
	y = y | (y >> 4);
	y = y | (y >> 8);
	y = y | (y >> 16);
	y = y + 1;
	return y;
}

int displayHelp(int argc, char *argv[]) {
	if (
		(argc < 2) ||
		(strcmp(argv[1], "-h") == 0) ||
		(strcmp(argv[1], "-?") == 0) ||
		(strcmp(argv[1], "--help") == 0)
	) {
		printf("-------------------------------------------------------------------\n");
		printf("Nintendo(R) Virtual Boy(TM) ISX to ROM Image Converter\n");
		printf("Copyright (C) 2010 Amos \"RunnerPack\" Bieler\n");
		printf("-------------------------------------------------------------------\n");
		printf("Usage: vbcvtisx.exe <infile.isx> [<outfile.vb>]\n\n");
		printf("The .ISX file 'infile.isx' will be parsed and code sections will be\n");
		printf("copied into a Virtual Boy(tm) ROM image called 'outfile.vb' (or\n");
		printf("'infile.vb' if no output name is specified).\n\n");
		printf("Notes/Known limitations:\n");
		printf("1. USE AT YOUR OWN RISK! THE AUTHOR SHALL NOT BE HELD RESPONSIBLE\n");
		printf("   FOR DAMAGE RESULTING FROM THE USE OF OR INABILITY TO USE THIS\n");
		printf("   SOFTWARE. See license.txt for more information.\n");
		printf("2. This program overwrites existing files without warning.\n");
		printf("3. Extensive testing has not been done due to lack of input data.\n");
		printf("4. Nothing is done with debug information contained in the .ISX file.\n");
		printf("   It will show up as 'unknown' chunks, but should be otherwise ignored.\n");
		printf("5. Nintendo does not endorse this software.\n");
		printf("6. 'Virtual Boy' is a trademark and 'Nintendo' is a registered\n");
		printf("   trademark of Nintendo Co. LTD. Both are used without permission.\n\n");

		exit(0);
	}
}

int main(int argc, char *argv[]) {
	FILE *fin, *fout;				// File pointers
	char namebuff[PATH_MAX];		// In case the user only supplies an input filename
	char signature[4] = "ISX";		// File signature
	char sigbuff[4] = "   ";		// File signature input buffer
	u32 isxsize;					// Size of the input file
	u32 romsize;					// Size of the output file
	float version;					// Version number of the input file
	u32 i;							// Miscellaneous iterator
	int inbyte;						// Holds a byte taken from the ISX file
	s32 addr;						// The tag address value
	s32 size;						// The tag size value
	u32 outaddr;					// The current working output address
	u8* buffer;						// A buffer to hold the content of an ISX section

	// If necessary, explain the program to the user.
	displayHelp(argc, argv);

	// Make an output filename if one is not supplied
	if (argc < 3) {
		strcpy(namebuff, argv[1]);
		for (i = strlen(namebuff); (namebuff[i] != '.') && (i > 1); i--);
		namebuff[i++] = '.';
		namebuff[i++] = 'v';
		namebuff[i++] = 'b';
		namebuff[i  ] = '\0';
	} else {
		strcpy(namebuff, argv[2]);
	}

	// Open input file for binary reading
	fin = fopen(argv[1], "rb");
	if (fin == NULL) {
		printf("Error: Unable to open input file.\n");
		exit(1);
	}

	// First, check for the ISX header:
	fread(sigbuff, 3, 1, fin);
	if (strcmp(signature, sigbuff) != 0) {
		printf("Error: This does not appear to be an ISX file.\n");
		exit(1);
	}

	// Get ISX file's size
	fseek(fin, 0, SEEK_END);
	isxsize = ftell(fin);
	fseek(fin, 32, SEEK_SET);	// Note: skip the (32 byte) header

	// Calculate smallest needed ROM size (must be an integral power of two)
	romsize = ceilp2(isxsize);

	// Make a "blank FlashROM" (filled with 0xFFs)

	// Give user a status report
	printf("Generating %d byte (%d Mbit) ROM file %s from %s...", romsize, romsize >> 17, namebuff, argv[1]);

	// Open output file for binary writing
	fout = fopen(namebuff, "wb");
	if (fout == NULL) {
		printf("\nError: Unable to open output file.\n");
		exit(1);
	}

	// Allocate and fill the buffer
	buffer = (u8*)malloc(romsize);
	if (buffer == NULL) {
		printf("\nError: Could not allocate memory for the blank ROM buffer.\n");
		exit(1);
	}
	memset(buffer, 0xFF, romsize);

	// Write it to the file and release the memory
	fwrite(buffer, romsize, 1, fout);
	free(buffer);

	// Begin extraction
	for (i = ftell(fin); i < isxsize - 9; i = ftell(fin)) {

		// Get a byte from the ISX file
		inbyte = fgetc(fin);

		// Check for end-of-file condition
		if (inbyte == EOF) break;

		// Check for a section tag
		if (inbyte == 0x11) {

			// Get the address
			fread(&addr, sizeof(addr), 1, fin);

			// Get the size
			fread(&size, sizeof(size), 1, fin);

			// If the target address is inside the ROM range, write it there
			if ((addr >= 0x07000000) && (addr < 0x08000000)) {

				// Adjust the offset to fit the disk file
				outaddr = addr & (romsize - 1);

				// Move the output pointer to the correct location
				fseek(fout, outaddr, SEEK_SET);

			// If the target address is negative, it represents an offset
			// from the end; calculate the actual address and write there
			} else if (addr < 0) {

				// Adjust the offset to fit the disk file
				outaddr = romsize + addr;

				// Move the output pointer to the correct location
				fseek(fout, outaddr, SEEK_SET);

			// If the tag seems incorrect, warn the user
			} else {
				printf("\n*** Warning: Malformed address value '0x%08X' found at offset %d! ***\n", addr, i - 9);
			}

		// Warn the user of an unknown chunk identifier
		} else {
			printf("\n*** Warning: Ignoring unknown chunk type '0x%02X' at offset %d.\n", inbyte, i);
			printf("*** Chunk address: '0x%08X' Chunk size: %08d\n", addr, size);
		}

		// Write the data block described by the tag to the ROM:

		// 1. Make the buffer to hold it
		buffer = (u8*)malloc(size);
		if (buffer == NULL) {
			printf("\nError: Could not allocate memory for input buffer.\n");
			exit(1);
		}

		// 2. Grab the block from the ISX
		fread(buffer, size, 1, fin);

		// 3. Write it to the ROM
		fwrite(buffer, size, 1, fout);

		// 4. Release the buffer memory
		free(buffer);
	}

	printf("Done!\n");
	exit(0);
}
